package com.proofpoint.http.client.netty;

import com.google.common.base.Objects;
import com.google.common.base.Preconditions;
import com.google.common.util.concurrent.AbstractFuture;
import com.proofpoint.http.client.AsyncHttpClient.AsyncHttpResponseFuture;
import com.proofpoint.http.client.Request;
import com.proofpoint.http.client.RequestStats;
import com.proofpoint.http.client.ResponseHandler;
import com.proofpoint.log.Logger;
import com.proofpoint.units.Duration;
import org.jboss.netty.handler.codec.http.HttpResponse;

import java.util.concurrent.CancellationException;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;
import java.util.concurrent.atomic.AtomicReference;

public class NettyResponseFuture<T, E extends Exception>
        extends AbstractFuture<T>
        implements AsyncHttpResponseFuture<T, E>
{
    public enum NettyAsyncHttpState
    {
        WAITING_FOR_CONNECTION,
        SENDING_REQUEST,
        WAITING_FOR_RESPONSE,
        PROCESSING_RESPONSE,
        DONE,
        FAILED,
        CANCELED
    }

    private final long requestStart = System.nanoTime();
    private final AtomicReference<NettyAsyncHttpState> state = new AtomicReference<>(NettyAsyncHttpState.WAITING_FOR_CONNECTION);
    private final Request request;
    private final ResponseHandler<T, E> responseHandler;
    private final RequestStats stats;


    public NettyResponseFuture(Request request, ResponseHandler<T, E> responseHandler, RequestStats stats)
    {
        this.request = request;
        this.responseHandler = responseHandler;
        this.stats = stats;
    }

    public String getState()
    {
        return state.get().toString();
    }

    protected void setState(NettyAsyncHttpState state)
    {
        this.state.set(state);
    }

    private static final Logger log = Logger.get(NettyResponseFuture.class);

    @Override
    protected boolean setException(Throwable throwable)
    {
        if (state.get() == NettyAsyncHttpState.CANCELED) {
            return false;
        }

        if (throwable instanceof CancellationException) {
            state.set(NettyAsyncHttpState.CANCELED);
        } else {
            state.set(NettyAsyncHttpState.FAILED);
        }
        if (throwable == null) {
            throwable = new Throwable("Throwable is null");
            log.error(throwable, "Something is broken");
        }

        // todo change response handler interface to accept throwable
        Exception exception;
        if (throwable instanceof Exception) {
            exception = (Exception) throwable;
        } else {
            exception = new Exception(throwable.getMessage(), throwable);
        }

        // give handler a chance to rewrite exception, and let them know there was a problem
        try {
            E e = responseHandler.handleException(request, exception);
            if (e != null) {
                throwable = e;
            }
        }
        catch (Throwable t) {
            throwable = t;
        }

        return super.setException(throwable);
    }

    @Override
    public boolean cancel(boolean mayInterruptIfRunning)
    {
        // Currently, we do not cancel pending requests
        state.set(NettyAsyncHttpState.CANCELED);
        return super.cancel(mayInterruptIfRunning);
    }

    protected void completed(HttpResponse httpResponse)
    {
        if (state.get() == NettyAsyncHttpState.CANCELED) {
            return;
        }

        // this time will not include the data fetching portion of the response,
        // since the response is fully cached in memory at this point
        long responseStart = System.nanoTime();

        state.set(NettyAsyncHttpState.PROCESSING_RESPONSE);
        NettyResponse response = null;
        try {
            response = new NettyResponse(httpResponse);
            T value = responseHandler.handle(request, response);
            set(value);
            state.set(NettyAsyncHttpState.DONE);
        }
        catch (Exception e) {
            setException(new ExceptionFromResponseHandler(e));
        }
        finally {
            Duration responseProcessingTime = Duration.nanosSince(responseStart);
            Duration requestProcessingTime = new Duration(responseStart - requestStart, TimeUnit.NANOSECONDS);

            if (response != null) {
                stats.record(request.getMethod(),
                        response.getStatusCode(),
                        response.getBytesRead(),
                        response.getBytesRead(),
                        requestProcessingTime,
                        responseProcessingTime);
            }
        }
    }

    @Override
    public T checkedGet()
            throws E
    {
        try {
            return get();
        }
        catch (InterruptedException | CancellationException | ExecutionException e) {
            throw mapException(e);
        }
    }

    @Override
    public T checkedGet(long timeout, TimeUnit unit)
            throws TimeoutException, E
    {
        try {
            return get(timeout, unit);
        }
        catch (InterruptedException | CancellationException | ExecutionException e) {
            throw mapException(e);
        }
    }

    private E mapException(Exception e)
    {
        if (e instanceof InterruptedException) {
            Thread.currentThread().interrupt();
        }

        if (e instanceof ExecutionException) {
            Throwable cause = e.getCause();
            // Do not ask the handler to "handle" an exception it produced
            if (cause instanceof ExceptionFromResponseHandler) {
                try {
                    return (E) cause.getCause();
                }
                catch (ClassCastException classCastException) {
                    // this should never happen but generics suck so be safe
                    // handler will be notified of the same exception again below
                }
            }
            if (cause instanceof Exception) {
                e = (Exception) cause;
            }
        }
        return responseHandler.handleException(request, e);
    }

    @Override
    public String toString()
    {
        return Objects.toStringHelper(this)
                .add("requestStart", requestStart)
                .add("state", state)
                .add("request", request)
                .toString();
    }

    private static class ExceptionFromResponseHandler
            extends Exception
    {
        private ExceptionFromResponseHandler(Exception cause)
        {
            super(Preconditions.checkNotNull(cause, "cause is null"));
        }
    }
}
